import subprocess
from google.cloud import storage
import google.auth
from datetime import datetime, timedelta
import base64
import requests
import urllib
import sys
import os
import json


class GCS(object):
    """
    Helper class for uploading and download files from Google Cloud Storage
    """
    def __init__(self):
        self.client = storage.Client.create_anonymous_client()
        self.authenticated = False

    def download(self, bucket, source, destination = None):
        """
        Downloads a file from GCS, handling both public and private buckets.
        
        Parameters
        ----------
        bucket: str
            the bucket name
        source: str
            the file path within the bucket
        destination: str | None
            the file path to which we would like to download the file.
            default to source path within current working directory
        """
        bucket = self.client.bucket(bucket)
        blob = bucket.blob(source)
        if destination is None:
            destination = os.path.join(os.getcwd(), source)
        blob.download_to_filename(destination)

    def authenticate(self):
        """
        Get user's default credentials on this machine.

        Returns nothing, sets the values into object attributes
        """
        # Use default credentials from the gcloud environment
        try:
            credentials, project = google.auth.default()
        except google.auth.DefaultCredentialsError:
            gcloud_login()
            credentials, project = google.auth.default()

        # Use the credentials to create a storage client
        self.authenticated = True
        self.client = storage.Client(credentials=credentials)

    def check_auth(self):
        """
        Authenticates the user if not already authenticated.
        See the `authenticate` method.
        """
        if not self.authenticated:
            self.authenticate()

    def upload(self, bucket, source, destination):
        """
        Uploads a file to GCS
        
        Parameters
        ----------
        bucket: str
            the bucket name
        source: str
            the file path on the local machine
        destination: str
            the file path within the bucket to which we would like 
            to upload the file.
        """
        # Use default credentials from the gcloud environment
        self.check_auth()

        # Upload the file
        bucket = self.client.bucket(bucket)
        blob = bucket.blob(destination)
        blob.upload_from_filename(source)

        print(f"File {source} uploaded to {destination}.")

    def upload_string(self, bucket, contents, destination):
        """
        Uploads a string to GCS as a file
        
        Parameters
        ----------
        bucket: str
            the bucket name
        source: str
            the file path on the local machine
        destination: str
            the file path within the bucket to which we would like 
            to upload the file.
        """
        # Use default credentials from the gcloud environment
        self.check_auth()

        # Upload the file
        bucket = self.client.bucket(bucket)
        blob = bucket.blob(destination)
        print(contents)
        blob.upload_from_string(contents)

        print(f"Contents uploaded to {destination}.")


def gcloud_login():
    """
    Helper method to use the gcloud API to get user credentials
    Requires the gcloud CLI tool to work:
    https://cloud.google.com/sdk/docs/install

    This method should redirect the user to a browser where they can sign in
    to their Google Account and grant permissions to the gcloud CLI

    Once the user is authenticated, we can use access tokens from the gcloud CLI
    to do gcloud stuff on the user's behalf.
    """
    subprocess.run(
        ["gcloud", "auth", "application-default", "login"],
        shell=True
    )

CACHED_ACCESS_TOKEN = None
def get_access_token() -> str:
    """Gets an access token using the gcloud CLI tool
    We can then use this access token to perform gcloud operations
    on behalf of the user.

    We cache the access token for future use, as this process
    may make network requests which are slow.    

    Returns
    -------
    access_token : str
        an access token that can be used to perform google cloud requests.
    """
    if CACHED_ACCESS_TOKEN is None:
        result = subprocess.run(
            ["gcloud", "auth", "print-access-token"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            shell=True
        )
        if result.returncode != 0:
            # getting the access token failed, probably because the user
            # isn't logged in.  Log the user in, and then try again.
            gcloud_login()
            input('''
                please complete gcloud login with your browser 
                and then hit the any key to continue
            ''')
            return get_access_token()
        CACHED_ACCESS_TOKEN = result.stdout.strip()
    return CACHED_ACCESS_TOKEN




import unittest
import pynoddgcs.publish
import utils
import pynoddgcs

class TestSomething(unittest.TestCase):

    def test_something(self):
        self.assertEqual(1,1)

def load_tests(loader, tests, ignore):
    modules = [
        pynoddgcs.publish,
        pynoddgcs.connect
    ]
    for module in modules:
        tests = utils.doctests(module, tests)
    return tests

if __name__ == '__main__':
    unittest.main()



from google.cloud import storage
import pycocotools.coco
import os
import urllib.parse
import io
import json
from .connect import GCS

# https://storage.googleapis.com/nmfs_odp_pifsc/PIFSC/SOD/MOUSS/jpg/20161014_205317_2/20161014.205850.178.003437.jpg
GCS_ROOT = 'https://storage.googleapis.com'
class NODDCOCODataset(object):
    """
    Class for validating and uploading annotations in COCO format.
    
    Parameters
    ----------
    coco: pycocotools.coco.COCO
        a COCO-format file
    dataset_root: str
        the root url path for datasets, including the bucket name, etc.
        consider computing with `dataset_path` function.
    """

    def __init__(self, coco_file: str, dataset_root: str, bucket:str): 
        self.coco_file = coco_file
        self.coco = pycocotools.coco.COCO(coco_file)
        self.coco_root = os.path.split(self.coco_file)[0]
        self.bucket = bucket
        self.relative_gcs_path = dataset_root
        self.gcs = GCS()

    def compute_urls(self):
        """
        Compute and update the urls within this COCO file to reflect the
        expected location of the files within GCS
        """
        for i, image in self.coco.imgs.items():
            # discard the drive letter, if present
            file_name = os.path.splitdrive(image['file_name'])[1]
            splitfile = split_filename(file_name)
            image['coco_url'] = join_urlpath(
                GCS_ROOT,
                self.bucket,
                self.relative_gcs_path, 
                *splitfile
            )

    def unnest_filenames(self, file_separator = '_'):
        """
        pycocotools doesn't appreciated "nested" `file_name attributes.
        To this end, we replace the `file_name` attribute to be "unnested",
        by simply replacing instances of '/' with some other separator.

        Note that the URL will still potentially contain nested paths.
        This method is idempotent if the file_name attribute is unnested.

        Parameters
        ----------
        file_separator: str
            the string we use to replace '/' in the `file_name`
        """
        # COCO file_names should not be nested
        for i, image in self.coco.imgs.items():
            # discard the drive letter, if present
            file_name = os.path.splitdrive(image['file_name'])[1]
            splitfile = split_filename(file_name)
            image['file_name'] = file_separator.join(splitfile)

    def upload_images(self):
        """
        Upload the images in this COCO metadata file to GCS.
        The files should be located at the location specified with 
        the `file_name` attribute, either absolute or relative to 
        the location of the COCO file.
        """
        for i, image in self.coco.imgs.items():
            print(image['file_name'])
            # discard the drive letter, if present
            file_name = os.path.splitdrive(image['file_name'])[1]
            splitfile = split_filename(file_name)
            destination = join_urlpath(
                self.relative_gcs_path, *splitfile
            )
            if os.path.isabs(image['file_name']):
                source = image['file_name']
            else:
                source = os.path.join(self.coco_root, image['file_name'])
            self.gcs.upload(self.bucket, source, destination)
    
    def upload_coco(self):
        """
        Upload the COCO file, but adjusted so that file urls point to files
        in the GCS bucket.
        """
        self.compute_urls()
        self.unnest_filenames()
        newcoco = json.dumps(self.coco.dataset)
        destination = join_urlpath(
            self.relative_gcs_path, 'annotations.json'
        )
        self.gcs.upload_string(self.bucket, newcoco, destination)

    def upload(self):
        """
        Upload this dataset, first the images, then the adjusted COCO file.
        """
        print("uploading images")
        self.upload_images()
        print("uploading coco file")
        self.upload_coco()

def dataset_path(datasets_root, organization, project):
    """
    Get a url path for uploading/downloading GCS NODD data based on
    the bucket, organization, etc.

    ```
    >>> dataset_path('bar', 'pickles+fish', 'project')
    'bar/pickles%2Bfish/project'
    
    ```

    Parameters
    ----------
    datasets_root: str
        the root directory within the bucket where we are hosting datasets
    organization: str
        the first-level fixed-depth directory for organizing datasets
    project: str
        the second-level fixed-depth directory for organizing datasets
    """
    return join_urlpath(
        datasets_root, organization, project)

def join_urlpath(*paths):
    """
    Joins a bunch of strings into a slash-delimited url path.

    ``` 
    >>> join_urlpath('foo', 'bar', 'pickles+fish', 'project')
    'foo/bar/pickles%2Bfish/project'
    
    >>> join_urlpath('http://foo', 'bar', 'pickles+fish', 'project')
    'http://foo/bar/pickles%2Bfish/project'

    >>> join_urlpath('http://foo/bar', 'pickles+fish/project')
    'http://foo/bar/pickles%2Bfish/project'
    
    ```

    Parameters
    ----------
    *paths: list[str]
        a variable-length list of path elements to join

    Returns
    -------
    url: str
        the joined url
    """
    url = '/'.join(s.strip('/') for s in paths)
    return urllib.parse.quote(url, safe=':/')
        
def split_filename(filename):
    """
    Splits a filename into all of its component directory structure

    Does the same thing as `os.path.split`, but completely splits the 
    directory structure into all parts, instead of just two (head/tail)

    ```
    
    >>> split_filename('foo/bar/pickles')
    ['foo', 'bar', 'pickles']

    ```

    Parameters
    ----------
    filename: str
        A filename to split

    Returns
    -------
    paths: list[str]
        The completely split list of directories 
        (and possibly the terminating filename)
    """
    paths = []
    tail = "totally_arbitrary"
    while filename and tail:
        filename, tail = os.path.split(filename)
        if tail and tail != '.' and tail != '..':
            paths.append(tail)
    return paths[::-1]